import { Test, TestingModule } from '@nestjs/testing';
import { getRepositoryToken } from '@nestjs/typeorm';
import { Project } from '../projects/project.entity';
import { ReposService } from './repos.service';
import { ConfigModule } from '@nestjs/config';
import { readFile, rm } from 'fs/promises';
import { join } from 'path';
import configuration from '../commons/config/configuration';
import { NotFoundException } from '@nestjs/common';

const getTest1Project = () =>
  ({
    id: '1',
    sshUrl: 'ssh://gitea@git.ivanli.cc:7018/Fennec/test-1.git',
    name: 'test-1',
  } as Project);

describe('ReposService', () => {
  let service: ReposService;
  const repositoryMockFactory = jest.fn(() => ({
    findOneOrFail: jest.fn(
      (entity): Project => ({
        ...getTest1Project(),
        ...entity,
      }),
    ),
  }));
  afterEach(async () => {
    await rm(service.getWorkspaceRoot(getTest1Project()), {
      recursive: true,
    }).catch(() => undefined);
  });
  beforeEach(async () => {
    const module: TestingModule = await Test.createTestingModule({
      imports: [
        ConfigModule.forRoot({
          load: [configuration],
        }),
      ],
      providers: [
        ReposService,
        {
          provide: getRepositoryToken(Project),
          useFactory: repositoryMockFactory,
        },
      ],
    }).compile();

    service = module.get<ReposService>(ReposService);

    await rm(service.getWorkspaceRoot(getTest1Project()), {
      recursive: true,
    }).catch(() => undefined);
  });

  it('should be defined', () => {
    expect(service).toBeDefined();
  });
  it('getWorkspaceRoot', () => {
    expect(service.getWorkspaceRoot(getTest1Project())).toBeDefined();
  });
  describe('listLogs', () => {
    it('should be return logs', async () => {
      const result = await service.listLogs({ projectId: '1' });
      expect(result).toBeDefined();
    }, 20_000);
  });
  describe('listBranch', () => {
    it('should be return branches', async () => {
      const result = await service.listBranches({ projectId: '1' });
      expect(result).toBeDefined();
    }, 10_000);
  });
  describe('checkoutBranch', () => {
    it('should be checkout', async () => {
      await service.checkoutBranch(getTest1Project(), 'master');
      const filePath = join(
        service.getWorkspaceRoot(getTest1Project(), 'master'),
        'README.md',
      );
      const text = await readFile(filePath, { encoding: 'utf-8' });
      expect(text).toMatch(/Commit 1/gi);
    }, 30_000);
    it('multiplexing workspace', async () => {
      await service.checkoutBranch(getTest1Project(), 'master');
      await service.checkoutBranch(getTest1Project(), 'branch-a');
      await service.checkoutBranch(getTest1Project(), 'branch-b');
      const filePath = join(
        service.getWorkspaceRoot(getTest1Project(), 'branch-b'),
        'branch-b.md',
      );
      const text = await readFile(filePath, { encoding: 'utf-8' });
      expect(text).toMatch(/Commit branch b/gi);
    }, 30_000);
    it('nonexistent branch', async () => {
      return expect(
        service.checkoutBranch(getTest1Project(), 'nonexistent'),
      ).rejects.toBeInstanceOf(NotFoundException);
    }, 30_000);
    it('checkout the specified version', async () => {
      await service.checkoutBranch(getTest1Project(), 'master');
      const filePath = join(
        service.getWorkspaceRoot(getTest1Project(), 'master'),
        'README.md',
      );
      const text = await readFile(filePath, { encoding: 'utf-8' });
      expect(text).toMatch(/Commit 1/gi);
    }, 30_000);
  });

  describe('checkoutCommit', () => {
    it('should be checkout', async () => {
      await service.checkoutCommit(getTest1Project(), '498c782685');
      const filePath = join(
        service.getWorkspaceRoot(getTest1Project(), '498c782685'),
        'README.md',
      );
      const text = await readFile(filePath, { encoding: 'utf-8' });
      expect(text).toMatch(/Commit 1/gi);
    }, 20_000);
    it('should be checkout right commit', async () => {
      await service.checkoutCommit(getTest1Project(), '7f7123fe5b');
      const filePath = join(
        service.getWorkspaceRoot(getTest1Project(), '7f7123fe5b'),
        'README.md',
      );
      const text = await readFile(filePath, { encoding: 'utf-8' });
      expect(text).toMatch(/(?!Commit 1)/gi);
    }, 20_000);
  });
});
