import { Injectable, NotFoundException } from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { F_OK } from 'constants';
import { access, mkdir } from 'fs/promises';
import { join } from 'path';
import { gitP } from 'simple-git';
import { Repository } from 'typeorm';
import { Project } from '../projects/project.entity';
import { ListBranchesArgs } from './dtos/list-branches.args';
import { ListLogsArgs } from './dtos/list-logs.args';
import { ConfigService } from '@nestjs/config';

const DEFAULT_REMOTE_NAME = 'origin';
@Injectable()
export class ReposService {
  constructor(
    @InjectRepository(Project)
    private readonly projectRepository: Repository<Project>,
    private readonly configService: ConfigService,
  ) {}

  getWorkspaceRoot(project: Project, subDir = ''): string {
    return join(
      this.configService.get<string>('workspaces.root'),
      project.name,
      encodeURIComponent(subDir),
    );
  }

  async lockWorkspace(workspaceRoot: string) {
    // TODO:  获取锁，失败抛错。
  }

  async getGit(project: Project, subDir?: string) {
    const workspaceRoot = this.getWorkspaceRoot(project, subDir);
    await this.lockWorkspace(workspaceRoot);

    const firstInit = await access(workspaceRoot, F_OK)
      .then(() => false)
      .catch(async () => {
        await mkdir(workspaceRoot, { recursive: true });
        return true;
      });
    const git = gitP(workspaceRoot);
    if (firstInit) {
      await git.init();
      await git.addRemote(DEFAULT_REMOTE_NAME, project.sshUrl);
    }
    return git;
  }

  async listLogs(dto: ListLogsArgs) {
    const project = await this.projectRepository.findOneOrFail({
      id: dto.projectId,
    });
    const git = await this.getGit(project);
    await git.fetch();
    return await git.log({
      '--branches': dto.branch ?? '',
      '--remotes': DEFAULT_REMOTE_NAME,
    });
  }

  async listBranches(dto: ListBranchesArgs) {
    const project = await this.projectRepository.findOneOrFail({
      id: dto.projectId,
    });
    const git = await this.getGit(project);
    return git.branch();
  }

  async checkoutBranch(project: Project, branch: string) {
    const git = await this.getGit(project, branch);
    try {
      await git.fetch(DEFAULT_REMOTE_NAME, branch);
    } catch (err) {
      if (err.message.includes("couldn't find remote ref nonexistent")) {
        throw new NotFoundException(err.message);
      }
      throw err;
    }
    await git.checkout([
      '-B',
      branch,
      '--track',
      `${DEFAULT_REMOTE_NAME}/${branch}`,
    ]);
  }

  async checkoutCommit(project: Project, commitNumber: string) {
    const git = await this.getGit(project, commitNumber);
    try {
      await git.fetch(DEFAULT_REMOTE_NAME);
    } catch (err) {
      if (err.message.includes("couldn't find remote ref nonexistent")) {
        throw new NotFoundException(err.message);
      }
      throw err;
    }
    await git.checkout([commitNumber]);
  }
}
